@@ -1,7 +1,7 @@
 /* Functions for communicating with a remote tape drive.
 
-   Copyright 1988, 1992, 1994, 1996, 1997, 1999, 2000, 2001, 2004 Free
-   Software Foundation, Inc.
+   Copyright (C) 1988, 1992, 1994, 1996, 1997, 1999, 2000, 2001, 2004,
+   2005, 2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* The man page rmt(8) for /etc/rmt documents the remote mag tape protocol
    which rdump and rrestore use.  Unfortunately, the man page is *WRONG*.
@@ -33,6 +33,8 @@
    code, courtesy of Dan Kegel.  */
 
 #include "system.h"
+#include "system-ioctl.h"
+
 #include <safe-read.h>
 #include <full-write.h>
 
@@ -340,9 +342,7 @@ encode_oflag (char *buf, int oflag)
 #ifdef O_NOCTTY
   if (oflag & O_NOCTTY) strcat (buf, "|O_NOCTTY");
 #endif
-#ifdef O_NONBLOCK
   if (oflag & O_NONBLOCK) strcat (buf, "|O_NONBLOCK");
-#endif
 #ifdef O_RSYNC
   if (oflag & O_RSYNC) strcat (buf, "|O_RSYNC");
 #endif
@@ -358,7 +358,7 @@ encode_oflag (char *buf, int oflag)
    remote pipe number plus BIAS.  REMOTE_SHELL may be overridden.  On
    error, return -1.  */
 int
-rmt_open__ (const char *file_name, int open_mode, int bias, 
+rmt_open__ (const char *file_name, int open_mode, int bias,
             const char *remote_shell)
 {
   int remote_pipe_number;	/* pseudo, biased file descriptor */
