@@ -55,8 +55,7 @@
    It must hold enough for any integer, possibly with a sign.  */
 #define	STRING_SIZE (UINTMAX_STRSIZE_BOUND + 1)
 
-/* Name of executing program.  */
-const char *program_name;
+char *program_name;
 
 /* File descriptor of the tape device, or negative if none open.  */
 static int tape = -1;
